#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sslauth.h"

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/rfcomm.h>

#include <obexftp/client.h>
#include <cobexbfb/cobex_bfb.h>


/*
static void print_dev_info(struct rfcomm_dev_info *di)
{
	char src[18], dst[18], addr[40];

	ba2str(&di->src, src); ba2str(&di->dst, dst);

	if (bacmp(&di->src, BDADDR_ANY) == 0)
		sprintf(addr, "%s", dst);
	else
		sprintf(addr, "%s -> %s", src, dst);

	printf("rfcomm%d: %s channel %d\n",
		di->id, addr, di->channel);
}
*/

static void info_cb(int event, const char *msg, int len, void *data) {
    // do nothing, here because still obexftp docs declares info_cb can be null, in
    // reality this causes segmentation fault!
}

int bluetooth_client(const char *tty,const char *tmp_dir) {
    obexftp_client_t *cli = NULL;
    obex_ctrans_t *ctrans = NULL;
    char lfile[MAX_FILENAME_SIZE+sizeof(spool_dir)+8], *lpos;
    char stmp[MAX_LINE_SIZE], *spos, *send;
    FILE *f;
    int res=0;

    strcpy(lfile,tmp_dir); lpos=lfile+strlen(lfile); *lpos='/'; lpos++; *lpos=0;
        
    ctrans = cobex_ctrans (tty);
    cli = obexftp_cli_open (info_cb, ctrans, NULL);
    if (!cli) {
#ifdef _DEBUG
	_pam_log(LOG_NOTICE,"Error opening obexftp-client connection over %s!",tty);
#endif
	if (ctrans) cobex_free(ctrans);	
	return 0;
    }

    if (obexftp_cli_connect (cli) < 0) {
#ifdef _DEBUG
	_pam_log(LOG_NOTICE,"Error connecting client over %s!",tty);
#endif
	obexftp_cli_close(cli);
	if (ctrans) cobex_free(ctrans);	
	return 0;
    }
    
    strcpy(lpos,"list");
    obexftp_list(cli,lfile,bt_path);
    f=fopen(lfile,"r");
    if (!f) {
#ifdef _DEBUG
	_pam_log(LOG_NOTICE,"Can't get list file over %s!",tty);
#endif
	obexftp_cli_disconnect (cli);
	obexftp_cli_close(cli);
        if (ctrans) cobex_free(ctrans);	
	return 0;
    }

    if (obexftp_setpath(cli, bt_path)>=0) {
	while(!feof(f)) {
	    fgets(stmp,MAX_LINE_SIZE-1,f);
	    for(spos=stmp;((*spos==' ')||(*spos=='\t'));spos++);
	    if (strncmp(spos,"<file name=\"",12)==0) {
		spos+=12;
	        send=strchr(spos,'"');
		if (!send) continue;
	        *send=0;
		send-=4;
	        if (send<=spos) continue;
		if ((strcmp(send,".crt"))&&(strcmp(send,".key"))) continue;
	        strcpy(lpos,spos);
		if (obexftp_get(cli, lfile, spos)>=0) res++;
		else {
#ifdef _DEBUG
		    _pam_log(LOG_WARNING,"Error downloading file %s over %s!",spos,tty);
#endif
		}
	    }
	}
    } else {
#ifdef _DEBUG
	_pam_log(LOG_WARNING,"Error changing to directory %s over %s!",bt_path,tty);
#endif
    }
    fclose(f);

    obexftp_cli_disconnect (cli);
    obexftp_cli_close(cli);
    if (ctrans) cobex_free(ctrans);	
    return res;
}


static int bluetooth(const char *tmp_dir) {
    int ctl;
    struct rfcomm_dev_list_req *dl;
    struct rfcomm_dev_info *di;
    char device[sizeof(rfcom_dir)+5];
    int i;
    int res=0;

    if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
	_pam_log(LOG_ERR,"Can't open RFCOMM control socket");
	return -1;
    }

    dl = malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(*di));
    if (!dl) {
    	_pam_log(LOG_ERR,"Can't allocate memory");
	close(ctl);
	return -1;
    }

    dl->dev_num = RFCOMM_MAX_DEV;
    di = dl->dev_info;

    if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
	_pam_log(LOG_ERR,"Can't get device list");
	free(dl);
	close(ctl);
	return -1;
    }

    for (i = 0; i < dl->dev_num; i++) {
	sprintf(device,rfcom_dir,(di+i)->id);
	res+=bluetooth_client(device,tmp_dir);
    }
    free(dl);
    close(ctl);
    if (res<2) return 1;
    return 0;
}

void BluetoothFree(char *tmp_dir) {
    char tmp_file[sizeof(spool_dir)+MAX_FILENAME_SIZE+8];

    DIR *dir;
    struct dirent *de;
    
    if (!tmp_dir) return;
    
    dir = opendir(tmp_dir);
    if (dir) {
	while ((de=readdir(dir))!=NULL) {
	    if ((strcmp(de->d_name,"."))&&(strcmp(de->d_name,".."))) {
		if (strlen(de->d_name)<MAX_FILENAME_SIZE) {
		    sprintf(tmp_file,"%s/%s",tmp_dir,de->d_name);
		    remove(tmp_file);
		} else {
		    _pam_log(LOG_ERR,"Invalid file detected in temporary directory, can't remove!");
		}
	    }
	}
    } else {
	_pam_log(LOG_ERR,"Error opening temporary directory!");
    }
    
    rmdir(tmp_dir);
}

char *BluetoothInit() {
    struct stat spool_stat;
    char *tmp_dir;

    if (stat(spool_dir,&spool_stat)!=0) {
	_pam_log(LOG_ERR,"Can't stat spool directory!");
	return NULL;
    }
    
    if (!S_ISDIR(spool_stat.st_mode)) {
	_pam_log(LOG_ERR,"Spool directory isn't directory!");
	return NULL;
    }
    
    if (spool_stat.st_mode&(S_IRWXG|S_IRWXO)) {
	_pam_log(LOG_ERR,"Spool directory have insecure permissions. Set 0700!");
	return NULL;
    }
    
    tmp_dir=(char*)malloc(sizeof(spool_dir)+6);
    if (!tmp_dir) {
	_pam_log(LOG_ERR,"Can't allocate memrory!");
	return NULL;
    }
    strcpy(tmp_dir,spool_dir "XXXXXX");
    
    if (!mkdtemp(tmp_dir)) {
	_pam_log(LOG_ERR,"Error creating temp directory!");
	free(tmp_dir);
	return NULL;
    }
    
    if (bluetooth(tmp_dir)) {
	BluetoothFree(tmp_dir);
	return NULL;
    }
    return tmp_dir;
}

/*
main() {
    char *tmp_dir;
    tmp_dir=BluetoothInit();
    BluetoothFree(tmp_dir);
}
*/
